import numpy as np
import editdistance

from numba import jit

#TODO add more

def makeLn(n):
    return lambda a, b: np.linalg.norm(a - b, ord=n)

def editDistance(a, b):
    return editdistance.eval(a, b)

@jit(nopython=True)
def earthMoverDistance(u, v):
    """
    Compute the first Wasserstein distance between two 1D distributions.
    :param u, v: (Array-like) Values observed in the (empirical) distribution.
    :return: (float) The computed distance between the distributions.
    """
    # Calculate the CDFs of u and v
    u_cumweights = np.concatenate((np.array([0.0]), np.cumsum(u)))
    u_cdf = u_cumweights / u_cumweights[-1]

    v_cumweights = np.concatenate((np.array([0.0]), np.cumsum(v)))
    v_cdf = v_cumweights / v_cumweights[-1]

    return np.sum(np.abs(u_cdf - v_cdf))



"""
Implementation of Algorithm 3 from [Near-optimal sample compression for nearest neighbors](https://www.cs.bgu.ac.il/~karyeh/condense-journal.pdf)

Epsilon Nets refer to https://en.wikipedia.org/wiki/Delone_set

variable names sadly avoid normal convention to correspond to the paper notations
"""

import numpy as np

from math  import floor
from numba import jit, prange

@jit(nopython=True)
def greedyConstructEpsilonNetWithGram(points, gram, epsilon):
    """
    Construct an Epsilon net in a greedy fashion.

    :param points: points to construct a net from
    :param gram: gram matrix of the points (for some metric)
    :param epsilon: epsilon parameter

    :return: the points chosen for the net and their indices (as an indicator vector)
    """
    idx = np.random.randint(0, len(points))

    netGram = np.full_like(gram, np.inf)
    taken   = np.full(len(points), False)

    netGram[idx] = gram[idx]
    taken[idx]   = True

    for i in range(len(points)):
        if np.min(netGram[:, i]) >= epsilon:
            netGram[i] = gram[i]
            taken[i]   = True

    return points[taken], taken

@jit(nopython=True)
def buildLevel(p, i, radius, gram, S, N, P, C):
    """
    Builds a level in the net hierarchy.
    (content of the loop of line 6 in the algorithm 3's pseudo-code)

    :param p: starting point
    :param i: current level in the hierarchy
    :param radius: radius for epsilon net
    :param gram: gram matrix of all the points
    :param S: the entire hierarchy
    :param N: neighbors
    :param P: parents
    :param C: covers
    """
    T = [e for l in [list(C[r, i - 1]) for x in P[p, i] for r in N[x, i]] for e in l] #TODO simplify or explain

    for r in T:
        if gram[r, p] < radius:
            P[p, i - 1] = {r}
            return

    S[i - 1].add(p)
    N[(p, i - 1)].add(p)
    [C[r, i - 1].add(p) for r in P[p, i]]

    for r in T:
        if gram[r, p] < 4 * radius:
            N[p, i - 1].add(r)
            N[r, i - 1].add(p)

# @jit(nopython=True)
def optimizedBuildLevel(p, i, radius, gram, S, N, P, C):
    """
    Optimized version of :func:buildLevel

    :param p: starting point
    :param i: current level in the hierarchy
    :param radius: radius for epsilon net
    :param gram: gram matrix of all the points
    :param S: the entire hierarchy
    :param N: neighbors
    :param P: parents
    :param C: covers
    """
    _P = P[p, i]
    if np.any(_P):
        _N = np.any(N[_P, i], axis=0)
        if np.any(_N):
            _C = np.any(C[_N, i + 1], axis=0)
            if np.any(_C):
                T = np.squeeze(np.argwhere(_C), axis=0)

                tGram = gram[p, T]
                j = np.argmin(tGram, axis=0)  # TODO ensure axis 0
                if tGram[j] < radius:
                    P[p, i + 1, T[j]] = True
                    return

                valid = [r for r in T if gram[p, r] < 4 * radius]
                N[p, i + 1, valid] = True
                N[valid, i + 1, p] = True

    S[i + 1, p]    = True
    N[p, i + 1, p] = True
    C[P[p, i], i + 1] |= P[p, i]

def hieracConstructEpsilonNet(points, gram, epsilon):
    """
    Construct an Epsilon net in a hierarchical fashion.
    Note: the resulting net has a radius 2^floor(log_2(epsilon))

    :param points: points to construct a net from
    :param gram: gram matrix of the points (for some metric)
    :param epsilon: epsilon parameter

    :return: the points chosen for the net and their indices (as an indicator vector)
    """
    lowestLvl = int(floor(np.log2(epsilon)))
    n         = len(points)
    levels    = range(1, lowestLvl - 1, -1)

    #arbitrary starting point
    startIdx = np.random.randint(0, n)

    #init S - nets
    S = {i: set() for i in levels}
    S[levels[0]].add(startIdx)

    #init P - parents
    P = {(p, i): set() for p in range(n) for i in levels}
    for p in range(n):
        P[p, levels[0]] = {startIdx}

    #init N - neighbors
    N = {(p, i): set() for p in range(n) for i in levels}
    N[startIdx, levels[0]] = {startIdx}

    #init C - covered
    C = {(p, i): set() for p in range(n) for i in levels}

    for i in levels[:-1]:
        radius = pow(2, i - 1)
        for p in S[i]:
            buildLevel(p, i, radius, gram, S, N, P, C)
        for p in set(range(n)) - S[i]:
            buildLevel(p, i, radius, gram, S, N, P, C)

    # guaranteed to by an e-net of at least epsilon
    return points[list(S[lowestLvl])], list(S[lowestLvl])

def optimizedHieracConstructEpsilonNet(points, gram, epsilon):
    """
    An optimized version of :func:hieracConstructEpsilonNet

    :param points: points to construct a net from
    :param gram: gram matrix of the points (for some metric)
    :param epsilon: epsilon parameter

    :return: the points chosen for the net and their indices (as an indicator vector)
    """
    lowestLvl = int(floor(np.log2(epsilon)))
    levels    = range(1, lowestLvl - 1, -1)

    n = len(points)
    l = len(levels)

    #arbitrary starting point
    startIdx = np.random.randint(0, n)

    Svec = np.zeros([l, n], dtype=np.bool)
    Nvec = np.zeros([n, l, n], dtype=np.bool)
    Pvec = np.zeros([n, l, n], dtype=np.bool)
    Cvec = np.zeros([n, l, n], dtype=np.bool)

    Svec[0, startIdx] = True
    Pvec[:, 0, startIdx] = True
    Nvec[startIdx, 0, startIdx] = True

    for j, i in enumerate(levels[:-1]):
        radius = pow(2, i - 1)
        for p in [k for k, x in enumerate(Svec[j,:]) if x]:
            optimizedBuildLevel(p, j, radius, gram, Svec, Nvec, Pvec, Cvec)
        for p in [k for k, x in enumerate(Svec[j,:]) if not x]:
            optimizedBuildLevel(p, j, radius, gram, Svec, Nvec, Pvec, Cvec)

    # guaranteed to by an e-net of at least epsilon
    netIndices = [i for i, x in enumerate(Svec[lowestLvl,:]) if x]
    idxs = np.zeros(n)
    idxs[netIndices] = True

    return points[netIndices], idxs



import logging
import numpy as np
import multiprocessing as mp

from time                     import time
from tqdm                     import tqdm
from tempfile                 import NamedTemporaryFile
from sklearn.neighbors        import KNeighborsClassifier
from sklearn.neighbors.base   import VALID_METRICS
from sklearn.metrics.pairwise import pairwise_distances


import ksu.Metrics
from ksu.epsilon_net.EpsilonNet import greedyConstructEpsilonNetWithGram, \
                                       optimizedHieracConstructEpsilonNet
from ksu.Utils import computeGammaSet, \
                      optimizedComputeLabels, \
                      optimizedComputeAlpha, \
                      computeQ, \
                      configLogger

METRICS = {v:v for v in VALID_METRICS['brute'] if v != 'precomputed'}
METRICS['EditDistance'] = ksu.Metrics.editDistance
METRICS['EarthMover']   = ksu.Metrics.earthMoverDistance

sharedLargeArrays = {}


def initWorker(X, Xshape, Y, Yshape, gram, gramShape):
    global sharedLargeArrays

    sharedLargeArrays['X'] = X
    sharedLargeArrays['X.shape'] = Xshape
    sharedLargeArrays['Y'] = Y
    sharedLargeArrays['Y.shape'] = Yshape
    sharedLargeArrays['gram'] = gram
    sharedLargeArrays['gram.shape'] = gramShape


def constructGammaNet(Xs, gram, gamma, prune=False, greedy=True):
    """
    Construct an epsilon net for parameter gamma

    :param Xs: points
    :param gram: gram matrix of the points
    :param gamma: net parameter
    :param prune: whether to prune the net after construction
    :param greedy: whether to build the net greedily or with an hierarchical strategy

    :return: the chosen points and their indices
    """
    if greedy:
        chosenXs, chosen = greedyConstructEpsilonNetWithGram(Xs, gram, gamma)
    else:
        chosenXs, chosen = optimizedHieracConstructEpsilonNet(Xs, gram, gamma)

    if prune:
        pass # TODO should we also implement this?

    return chosenXs, chosen.astype(np.bool)


def compressDataWorker(i, gammaSet, tmpFile, delta, minC, maxC, greedy):
    # load shared arrays
    Xs   = np.frombuffer(sharedLargeArrays['X'],    dtype=np.float32).reshape(sharedLargeArrays['X.shape'])
    Ys   = np.frombuffer(sharedLargeArrays['Y'],    dtype=np.int32)  .reshape(sharedLargeArrays['Y.shape'])
    gram = np.frombuffer(sharedLargeArrays['gram'], dtype=np.float32).reshape(sharedLargeArrays['gram.shape'])

    n           = len(Xs)
    numClasses  = len(np.unique(Ys))
    bestGamma   = 0.0
    bestCompres = 0.0
    qMin        = float(np.inf)
    chosenXs    = np.empty(0)
    chosenYs    = np.empty(0)
    chosenIdxs  = np.empty(0)
    logger      = logging.getLogger('KSU')


    logger.debug('Choosing from {} gammas'.format(len(gammaSet)))
    for gamma in tqdm(gammaSet):
        tStart = time()
        gammaXs, gammaIdxs = constructGammaNet(Xs, gram, gamma, greedy=greedy)
        compression = float(np.sum(gammaIdxs)) / n
        logger.debug('Gamma: {g}, net construction took {t:.3f}s, compression: {c}'.format(
            g=gamma,
            t=time() - tStart,
            c=compression))

        if not (minC < compression < maxC):
            continue  # don't bother compressing if not within limits

        if len(gammaXs) < numClasses:
            logger.debug(
                'Gamma: {g}, compressed set smaller than number of classes ({cc} vs {c})\n'
                'no use building a classifier that will never classify some classes'.format(
                    g=gamma,
                    cc=len(gammaXs),
                    c=numClasses))
            break # heuristic: gammas are increasing, so we might as well stop here

        tStart  = time()
        gammaYs = optimizedComputeLabels(gammaIdxs, gram, Ys)
        logger.debug('Gamma: {g}, label voting took {t:.3f}s'.format(
            g=gamma,
            t=time() - tStart))

        tStart = time()
        alpha  = optimizedComputeAlpha(gammaYs, Ys, gram[gammaIdxs])
        logger.debug('Gamma: {g}, error approximation took {t:.3f}s, error: {a}'.format(
            g=gamma,
            t=time() - tStart,
            a=alpha))

        m = len(gammaXs)
        q = computeQ(n, 2 * m, alpha, delta)

        if q < qMin:
            logger.info(
                'Gamma: {g} achieved lowest q so far: {q}, for compression {c}, and empirical error {a}'.format(
                    g=gamma,
                    q=q,
                    c=compression,
                    a=alpha))

            qMin        = q
            bestGamma   = gamma
            chosenXs    = gammaXs
            chosenYs    = gammaYs
            chosenIdxs  = gammaIdxs
            bestCompres = compression

    if qMin < float(np.inf):
        logger.info('Chosen best gamma: {g}, which achieved q: {q}, and compression: {c}'.format(
            g=bestGamma,
            q=qMin,
            c=bestCompres))
    else:
        logger.info('No good gamma found between gammas: [{min} ... {max}]'.format(
            min=gammaSet.min(),
            max=gammaSet.max()))

    np.savez_compressed(tmpFile, X=chosenXs, Y=chosenYs, idxs=chosenIdxs)
    return qMin, i, bestCompres, bestGamma


class KSU(object):

    def __init__(self, Xs, Ys, metric, gram=None, prune=False, logLevel=logging.CRITICAL, n_jobs=1):
        self.metric      = metric
        self.n_jobs      = n_jobs
        self.chosenXs    = None
        self.chosenYs    = None
        self.chosenIdxs  = None
        self.compression = None
        self.numClasses  = len(np.unique(Ys))
        self.prune       = prune # unused since pruning is not implemented yet

        self.logger = logging.getLogger('KSU')
        configLogger(self.logger, logLevel)

        if isinstance(metric, str) and metric not in METRICS.keys():
            raise RuntimeError(
                '"{m}" is not a built-in metric. use one of'
                '{ms}'
                'or provide your own custom metric as a callable'.format(
                    m=metric,
                    ms=METRICS.keys()))

        if gram is None:
            self.logger.info('Computing Gram matrix...')
            tStartGram = time()
            gram  = pairwise_distances(Xs, metric=self.metric, n_jobs=self.n_jobs)
            self.logger.debug('Gram computation took {:.3f}s'.format(time() - tStartGram))

        gram /= gram.max()

        # create shared arrays
        self.Xs   = mp.RawArray('f', int(np.prod(Xs.shape)))
        self.Ys   = mp.RawArray('i', int(np.prod(Ys.shape)))
        self.gram = mp.RawArray('f', int(np.prod(gram.shape)))

        np.copyto(np.frombuffer(self.Xs,   dtype=np.float32).reshape(Xs.shape),   Xs.astype(np.float32))
        np.copyto(np.frombuffer(self.Ys,   dtype=np.int32)  .reshape(Ys.shape),   Ys.astype(np.int32))
        np.copyto(np.frombuffer(self.gram, dtype=np.float32).reshape(gram.shape), gram.astype(np.float32))

        self.XsShape   = Xs.shape
        self.YsShape   = Ys.shape
        self.gramShape = gram.shape


    def getCompressedSet(self):
        """
        Getter for compressed set

        :return: the compressed set and its labels

        :raise: :class:RuntimeError if :func:KSU.KSU.compressData was not run before
        """
        if self.chosenXs is None:
            raise RuntimeError('getCompressedSet - you must run KSU.compressData first')

        return self.chosenXs, self.chosenYs


    def getCompression(self):
        """
        Getter for compression ratio

        :return: the compression ratio

        :raise: :class:RuntimeError if :func:KSU.KSU.compressData was not run before
        """
        if self.compression is None:
            raise RuntimeError('getCompression - you must run KSU.compressData first')

        return self.compression


    def getClassifier(self):
        """
        Getter for 1-NN classifier fitted on the compressed set

        :return: an :mod:sklearn.KNeighborsClassifier instance

        :raise: :class:RuntimeError if :func:KSU.KSU.compressData was not run before
        """
        if self.chosenXs is None:
            raise RuntimeError('getClassifier - you must run KSU.compressData first')

        h = KNeighborsClassifier(n_neighbors=1, metric=self.metric, algorithm='auto', n_jobs=self.n_jobs)
        h.fit(self.chosenXs, self.chosenYs)

        return h


    def compressData(self, delta=0.1, minCompress=0.01, maxCompress=0.1, greedy=True, stride=200):
        """
        Run the KSU algorithm to compress the dataset

        :param delta: confidence for error upper bound
        :param minCompress: minimum compression ratio
        :param maxCompress: maximum compression ratio
        :param greedy: whether to use greedy or hierarchical strategy for net construction
        :param stride: how many gammas to skip between each iteration (similar gammas will produce similar nets)
        """
        gammaSet = computeGammaSet(self.gram, stride=stride).astype(np.float32) # TODO add heuristic to throw away gammas that won't satisfy the compression limits
        numProcs = self.n_jobs if self.n_jobs > 0 else mp.cpu_count() + 1 + self.n_jobs
        self.logger.info('using {n} cores to process {g} gammas'.format(n=numProcs, g=len(gammaSet)))

        if numProcs == 1:
            numJobs = 1
        else:
            numJobs = 4 * numProcs

        if gammaSet.shape[0] % numJobs > 0:
            extra = -(gammaSet.shape[0] % -numJobs)
            padding = np.ones([extra], dtype=gammaSet.dtype)
            gammaSet = np.concatenate([gammaSet, padding])

        # mp.log_to_stderr(logging.DEBUG) # uncomment to debug concurrency

        tmpFiles  = [NamedTemporaryFile() for _ in range(numJobs)]
        gammaSets = np.reshape(gammaSet, [numJobs, -1])
        pool      = mp.Pool(processes=numProcs,
                            initializer=initWorker,
                            initargs=(self.Xs, self.XsShape, self.Ys, self.YsShape, self.gram, self.gramShape,))
        results   = [pool.apply_async(func=compressDataWorker,
                                      args=(i, gammaSets[i], tmpFiles[i].name + '.npz', delta, minCompress, maxCompress, greedy,),)
                     for i in range(numJobs)]
        pool.close()
        pool.join()

        results = sorted([r.get() for r in results], key=lambda r: r[0])  # sorted by qMin
        qMin, i, self.compression, bestGamma = results[0]
        tmpFile = tmpFiles[i] if i is not None else 0

        if qMin == float(np.inf):
            self.logger.critical('No gamma was chosen! check logs')
            return

        chosen          = np.load(tmpFile.name + '.npz', allow_pickle=True)
        self.chosenXs   = chosen['X']
        self.chosenYs   = chosen['Y']
        self.chosenIdxs = chosen['idxs']

        self.logger.info('Chosen best gamma: {g}, which achieved q: {q}, and compression: {c}'.format(
            g=bestGamma,
            q=qMin,
            c=self.compression))




import unittest
import sys
import pytest
import numpy as np

from sklearn.metrics.pairwise import pairwise_distances

from ksu.Utils import computeLabels, optimizedComputeLabels


class TestComputeGram(unittest.TestCase):

    @staticmethod
    def suite():
        test_suite = unittest.TestSuite()
        test_suite.addTest(unittest.makeSuite(TestComputeGram))
        return test_suite

    def empty_elements(self):
        self.fail()

    def empty_metric(self):
        self.fail()

    def first(self):
        self.fail()

xs = np.array([[0, 0],
               [0, 1],
               [1, 0],
               [1, 1],
               [2, 1],
               [3, 1],
               [2, 2],
               [3, 2]])
ys = np.array([2, 0, 1, 2, 0, 1, 0, 2])

slice = [1, 6]
subXs = xs[slice]

def testComputeLabels():
    assert computeLabels(subXs, xs, ys, 'l2') == [2, 0]

def testOptComputeLabels():
    gram = pairwise_distances(xs, metric='l2')
    assert optimizedComputeLabels(subXs, slice, xs, ys, gram) == [2, 0]


if __name__ == "__main__":
    # So you can run tests from this module individually.
    sys.exit(unittest.main())



import time
import logging
import sys
import numpy as np

from sklearn.neighbors import KNeighborsClassifier
from collections       import Counter
from math              import sqrt
from tqdm              import tqdm
from numba             import jit, prange


LOGGER_FORMAT = '%(asctime)s [%(levelname)s] PID %(process)d: %(module)s - %(message)s'
READABLE_TIME = '%Y-%m-%d_%H.%M.%S'


class TqdmHandler(logging.Handler):
    def __init__ (self, level=logging.NOTSET):
        super(self.__class__, self).__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.write(msg)
        except KeyboardInterrupt:
            pass
        except:
            self.handleError(record)

class TqdmStream(object):
    @classmethod
    def write(cls, msg):
        tqdm.write(msg, end='')

def getDateTime():
    return time.strftime(READABLE_TIME)

def parseInputData(dataPath):
    nodes = np.load(dataPath)
    try:
        data = {node: nodes[node] for node in ['X', 'Y']}
    except KeyError:
        raise RuntimeError('file at {p} does not contain the nodes "X" "Y"'.format(p=dataPath))

    return data

def computeGram(elements, dist): #unused
    n    = len(elements)
    gram = np.array((n, n))
    for i in range(n):
        for j in range(n - i):
            gram[i,j] = dist(elements[i], elements[j])

    lowTriIdxs       = np.tril_indices(n) #TODO make sure gram is upper triangular after the loop
    gram[lowTriIdxs] = gram.T[lowTriIdxs]

    return gram

@jit(nopython=True)
def computeQ(n, m, alpha, delta, w1=1, w2=1, w3=1):
    """
    Compute the parameter q that approximates an upper limit for the
    error of a 1-NN classifier based on the compressed set

    :param n: size of original set
    :param m: size of compressed set
    :param alpha: empirical error on the original set
    :param delta: level of confidence
    :param w1: first term weight
    :param w2: second term weight
    :param w3: third term weight

    :return: the approximation q
    """
    if m >= n:
        return float(np.inf)

    firstTerm  = (n * alpha) / (n - m)
    secondTerm = (m * np.log2(n) - np.log2(delta)) / (n - m)
    thirdTerm  = sqrt(((n * m * alpha * np.log2(n)) / (n - m) - np.log2(delta)) / (n - m))

    return w1 * firstTerm + w2 * secondTerm + w3 * thirdTerm

def computeLabels(gammaXs, Xs, Ys, metric, n_jobs=1): # unused
    """
    Compute the labels of the compressed set with a nearest neighbor majority vote

    :param gammaXs: the compressed set
    :param Xs: the original set
    :param Ys: the original labels
    :param metric: the distance metric
    :param n_jobs: number of cpus to employ (with scipy logic)

    :return: the labels of the compressed set
    """
    gammaN  = len(gammaXs)
    gammaYs = range(gammaN)
    h       = KNeighborsClassifier(n_neighbors=1, metric=metric, algorithm='auto', n_jobs=n_jobs)
    h.fit(gammaXs, gammaYs)
    groups      = [Counter() for _ in range(gammaN)]
    predictions = h.predict(Xs) # cluster id for each x (ids form gammaYs)
    [groups[label].update(Ys[np.where(predictions == label)]) for label in gammaYs] # count all the labels in the cluster

    return np.array([c.most_common(1)[0][0] for c in groups])


def optimizedComputeLabels(gammaIdxs, gram, Ys, neighbors=3, n_jobs=1):
    """
    An optimized version of :func:computeLabels
    Compute the labels of the compressed set with a nearest neighbor majority vote

    :param gammaIdxs: the compressed set indices
    :param gram: the distance matrix
    :param Ys: the original labels
    :param neighbors: number of neighbors to consider
    :param n_jobs: number of cpus to employ (with scipy logic)

    :return: the labels of the compressed set
    """
    h = KNeighborsClassifier(n_neighbors=neighbors, metric='precomputed', algorithm='auto', n_jobs=n_jobs)
    h.fit(gram, Ys)
    nearest = np.squeeze(h.kneighbors()[1]) # take only indices
    representers = majorityLabels(Ys[nearest])

    return representers[gammaIdxs]


def computeAlpha(gammaXs, gammaYs, Xs, Ys, metric):
    """
    Compute the empirical error of the classifier fitted to the compressed set
    on the original set

    :param gammaXs: compressed set points
    :param gammaYs: compressed set labels
    :param Xs: original set points
    :param Ys: original set labels
    :param metric: distance metric

    :return: the misclassification error
    """
    classifier = KNeighborsClassifier(n_neighbors=1, metric=metric, algorithm='auto', n_jobs=-1)
    classifier.fit(gammaXs, gammaYs)

    return classifier.score(Xs, Ys)


def optimizedComputeAlpha(gammaYs, Ys, gammaGram):
    """
    Optimized version of :func:computeAlpha

    :param gammaYs: compressed set labels
    :param Ys: original set labels
    :param gammaGram: rows of the original gram matrix corresponding to the compressed set

    :return: the misclassification error
    """
    nearest = np.argmin(gammaGram, axis=0) # nearest neighbors' indices in the compressed set
    missed  = Ys != gammaYs[nearest]

    return np.mean(missed)


def computeGammaSet(gram, stride=None):
    gammaSet = np.unique(gram)
    gammaSet = np.delete(gammaSet, 0)

    if stride is not None:
        gammaSet = gammaSet[::int(stride)]

    return gammaSet


def configLogger(logger, level):
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(fmt=LOGGER_FORMAT, datefmt=READABLE_TIME))
    handler.setLevel(level)
    logger.addHandler(handler)
    logger.setLevel(level)


@jit(nopython=True, nogil=True)
def majorityVote(arr):
    """
    if all elements are identical, return the first. otherwise return the majority
    """
    if arr[0] == arr.mean():
        return arr[0]

    elems  = np.unique(arr) # return_counts is not supported in numba version, this is still faster
    counts = np.array([np.sum(arr == i) for i in elems])
    argmax = np.argmax(counts)

    return elems[argmax]


@jit(nopython=True, nogil=True)
def majorityLabels(neighbors):
    ys = np.zeros(neighbors.shape[0], dtype=np.int32)
    for i in prange(neighbors.shape[0]):
        ys[i] = majorityVote(neighbors[i])

    return ys



import sys
import argparse
import logging
import numpy as np

from time                     import time
from sklearn.metrics.pairwise import pairwise_distances

from Utils   import parseInputData
from ksu.KSU import METRICS
from ksu.epsilon_net import EpsilonNet

def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Generate an epsilon net of the dataset')
    parser.add_argument('--data_in',       help='Path to input data file (in .npz format with a node named X)\n'
                                                'If another node named Y is present, the script will save the'
                                                'respective labels from Y',                                                    required=True)
    parser.add_argument('--data_out',      help='Path where output data will be saved',                                        required=True)
    parser.add_argument('--epsilon',       help='Required epsilon of net',                                                     required=True, type=float)
    parser.add_argument('--metric',        help='Metric to use (unless custom_metric is provided). {}'.format(METRICS.keys()), default='l2')
    parser.add_argument('--custom_metric', help='Absolute path to a directory (containing __init__.py) with a python file'
                                                'named Distance.py with a function named "dist(a, b)" that computes'
                                                'the distance between a and b by any metric of choice',                        default=None)
    parser.add_argument('--mode',          help='which constuction mode.\n'
                                                '"G" for greedy (faster, but bigger net), "H" for hierarchical',               default="G")
    parser.add_argument('--gram',          help='Path to a precomputed gram matrix (in .npz format with a node named gram)',   default=None)
    parser.add_argument('--log_level',     help='Logging level',                                                               default='INFO')

    args = parser.parse_args()

    logging.basicConfig(level=args.log_level, filename='e-net.log')
    logger = logging.getLogger('E-Net')
    logger.addHandler(logging.StreamHandler(sys.stdout))

    dataInPath   = args.data_in
    dataOutPath  = args.data_out
    gramPath     = args.gram
    metric       = args.metric
    epsilon      = args.epsilon
    mode         = args.mode
    customMetric = args.custom_metric

    if mode not in ['G', 'H']:
        raise RuntimeError('Mode {} is not supported. must be either of "G" for greedy (faster, but bigger net), "H" for hierarchical'.format(mode))

    if customMetric is not None:
        sys.path.append(customMetric)
        try:
            from Distance import dist  # this only looks like an error because we're importing from unknown path
            metric = dist
            logger.debug('Loaded dist function successfully')
        except:
            raise RuntimeError(
                'Could not import dist function from {p}'
                'make sure Distance.py and __init__.py exist in {p}'
                'and that Distance.py has a function dist(a, b)'.format(p=customMetric))
    else:
        if metric not in METRICS.keys():
            raise RuntimeError(
                '"{m}" is not a built-in metric. use one of'
                '{ms}'
                'or provide a custom metric with the --custom_metric argument'.format(
                    m=metric,
                    ms=METRICS.keys()))

    logger.info('Reading data...')
    data = np.load(dataInPath)
    try:
        Xs = data['X']
    except KeyError:
        raise RuntimeError('{} does not contain a node named X'.format(dataInPath))

    try:
        Ys = data['Y']
    except KeyError:
        Ys = None
        logger.debug('No labels')

    if gramPath is None:
        logger.info('Computing Gram matrix...')
        tStartGram = time()
        gram = pairwise_distances(Xs, metric=metric)
        logger.debug('Gram computation took {:.3f}s'.format(time() - tStartGram))
    else:
        logger.info('Loading gram...')
        gram = np.load(gramPath)['gram']

    gram = gram / np.max(gram)

    if mode == 'G':
        construct = EpsilonNet.greedyConstructEpsilonNetWithGram
    else:
        construct = EpsilonNet.hieracConstructEpsilonNet

    tStartNet   = time()
    net, idxs   = construct(Xs, gram, epsilon)
    compression = float(len(net)) / len(Xs)
    logger.info('Achieved {c} compression in time {t}s\n'
                'saving compressed set to {p}...'.format(
        c=compression,
        p=dataOutPath,
        t=time() - tStartNet))

    dataOut = {'X': net}
    if Ys is not None:
        dataOut = {'Y': Ys[idxs]}

    np.savez_compressed(dataOutPath, **dataOut)
    logger.info('Done')

if __name__ == '__main__' :
    sys.exit(main())



import sys
import argparse
import logging
import numpy as np

from ksu.Utils import parseInputData
from ksu.KSU   import KSU, METRICS

def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Generate a 1 nearest neighbors classifier fitted to a KSU compressed dataset')
    parser.add_argument('--data_in',         help='Path to input data file (in .npz format with 2 nodes named X and Y)',          required=True)
    parser.add_argument('--data_out',        help='Path where output data will be saved',                                         required=True)
    parser.add_argument('--metric',          help='Metric to use (unless custom_metric is provided)',                             default='l2', choices=METRICS.keys())
    parser.add_argument('--custom_metric',   help='Absolute path to a directory (containing __init__.py) with a python file'
                                                  'named Distance.py with a function named "dist(a, b)" that computes'
                                                  'the distance between a and b by any metric of choice',                         default=None)
    parser.add_argument('--gram',            help='Path to a precomputed gram matrix (in .npz format with a node named gram)',    default=None)
    parser.add_argument('--compress_limits', help='high,low compression ratio limits (comma separated, no spaces)',               default=None)
    parser.add_argument('--stride',          help='How many gammas to skip at a time (similar gammas will produce similar nets)', default=200, type=int)
    parser.add_argument('--delta',           help='Required confidence level',                                                    default=0.05, type=float)
    parser.add_argument('--mode',            help='which constuction mode.\n'
                                                  '"G" for greedy (faster, but bigger net), "H" for hierarchical',                default="G", choices=['G', 'H'])
    parser.add_argument('--num_procs',       help='Number of processes to use for computation',                                   default=1, type=int)
    parser.add_argument('--log_level',       help='Logging level',                                                                default=logging.CRITICAL)

    args = parser.parse_args()

    dataInPath   = args.data_in
    dataOutPath  = args.data_out
    gramPath     = args.gram
    metric       = args.metric
    delta        = args.delta
    mode         = args.mode
    customMetric = args.custom_metric
    compressLims = args.compress_limits
    stride       = args.stride
    logLevel     = args.log_level
    numProcs     = args.num_procs

    logging.basicConfig(level=logLevel, filename='ksu.log')
    logger = logging.getLogger('KSU')
    logger.addHandler(logging.StreamHandler(sys.stdout))

    greedy = mode == 'G'

    if customMetric is not None:
        sys.path.append(customMetric)
        try:
            from Distance import dist  # this only looks like an error because we're importing from an unknown path
            metric = dist
            logger.debug('Loaded dist function successfully')
        except:
            raise RuntimeError(
                'Could not import dist function from {p}'
                'make sure Distance.py and __init__.py exist in {p}'
                'and that Distance.py has a function dist(a, b)'.format(p=customMetric))
    else:
        if metric not in METRICS.keys():
            raise RuntimeError(
                '"{m}" is not a built-in metric. use one of'
                '{ms}'
                'or provide a custom metric with the --custom_metric argument'.format(
                    m=metric,
                    ms=METRICS.keys()))

    logger.info('Reading data...')
    data = parseInputData(dataInPath)
    gram = None
    if gramPath is not None:
        logger.info('Loading gram...')
        gram = np.load(gramPath)['gram']

    if compressLims is None:
        maxC = 1.0
        minC = 0.0
    else:
        ratios = compressLims.split(',')
        try:
            maxC = float(ratios[0])
            minC = float(ratios[1])
        except (IndexError, ValueError):
            raise RuntimeError('compress_limits shoud be two floats, comma separated, no spaces (e.g. "0.5,0.01").\ngiven {}'.format(compressLims))

        if maxC < minC:
            raise RuntimeError('compress_limits argument order is <high>,<low>')

    ksu = KSU(data['X'], data['Y'], metric, gram, logLevel=logLevel, n_jobs=1)
    ksu.compressData(delta=delta, maxCompress=maxC, minCompress=minC, greedy=greedy, stride=stride, numProcs=numProcs, logLevel=logLevel)
    Xs, Ys      = ksu.getCompressedSet()
    compression = ksu.getCompression()

    logger.info('Achieved {} compression, saving compressed set to {}'.format(compression, dataOutPath))
    np.savez_compressed(dataOutPath, X=Xs, Y=Ys)

    logger.info('Done')

if __name__ == '__main__' :
    sys.exit(main())



import pytest
import numpy as np
from sklearn.metrics.pairwise import pairwise_distances

from ksu.epsilon_net.EpsilonNet import greedyConstructEpsilonNetWithGram as e_net

x0   = np.array([0, 0])
x1   = np.array([0, 1])
x2   = np.array([1, 0])
x3   = np.array([1, 1])
xs   = np.vstack((x0, x1, x2, x3))
gram = pairwise_distances(xs, metric='l2')

@pytest.mark.parametrize('eps,shape', [(0.9, (4, 2)),
                                       (1.1, (2, 2)),
                                       (1.3, (2, 2)),
                                       (1.5, (1,2))])
def testGreedyConstruct(eps, shape):
    assert np.shape(e_net(xs, gram, eps)[0]) == shape
